"""
Ludolph: Monitoring Jabber Bot
Version: 0.4.0
Homepage: https://github.com/erigones/Ludolph
Copyright (C) 2012-2013 Erigones s. r. o.

See the LICENSE file for copying permission.
"""

__version__ = '0.4.0'
